﻿-- (c) eradicator a.k.a lossycrypt, 2017-2021, not seperately licensable
-- -------------------------------------------------------------------------- --

local const = {}

const.index = {

  -- "last used"
  -- Index order must be obeyed by de/encode function argument order.
  lu = {
    logistic_mode = 1,
    width         = 2,
    height        = 3,
    },

  -- Savedata.stockpiles[uid]
  pile = {
    container   = 1,
    render      = 2,
    infinity    = 3,
    
    uid         = 99, -- really needed?
    },
  
  -- Savedata.stockpiles[uid][const.index.pile.infinity]  
  infinity = {
    simple_stack = 1,
    half_count   = 2,
    stored_count = 3,
    total_count  = 4,
    render_id    = 5, -- number
    inventory    = 6,
    },
  
  render = {
    visual_mode = 1,
    renders     = 2, -- render id? simple entity?
    },
  
  }
  
  
const.name = {
  item           = 'er:stockpile',
  recipe         = 'er:stockpile',
  entity_prefix  = 'er:stockpile', -- must not contain lua string patterns
  blueprint      = 'er:stockpile-blueprint',
  }
  
const.gui = {
  name = {
    window_anchor = 'er:stockpile-anchor',
    },
  }
  
const.stockpile = {
  -- Increasing these settings can causes massive ram usage spikes
  -- during startup (16G+!).

  min_width      =  2,
  min_height     =  2,
  max_width      = 32,
  max_height     = 32,
  default_width  =  6,
  default_height =  6,
  
  max_area       = 16 * 32,
  
  default_logistic_mode       = 'container',
  slots_per_tile_intersection =           9,
  maximum_slot_count          =        5000,
  
  
  infinity = {
    ItemPrototypeFilter = { {filter = 'type', type = 'item'} },
    },
  
  -- DEV OVERRIDES
  -- max_area       =  6 * 12, -- faster dev load
  -- max_area       = 32 * 32, -- full range
  -- max_area       = 64 * 64, -- über range
  -- max_width      = 64,
  -- max_height     = 64,
  
  }

  
  
-- Determines gui icon order.
const.stockpile.logistic_modes = (function(r)
  local n = #r
  for i, this in ipairs(r) do
    this.prev, this.next = r[(i-2)%n+1], r[ i   %n+1]
    r[this.logistic_mode] = this
    end
  for i in ipairs(r) do r[i] = nil end
  return r end){
  {logistic_mode = 'container'       , sprite = 'entity/steel-chest'                    },
  {logistic_mode = 'passive-provider', sprite = 'entity/logistic-chest-passive-provider'},
  {logistic_mode = 'active-provider' , sprite = 'entity/logistic-chest-active-provider' },
  {logistic_mode = 'storage'         , sprite = 'entity/logistic-chest-storage'         },
  {logistic_mode = 'buffer'          , sprite = 'entity/logistic-chest-buffer'          },
  {logistic_mode = 'requester'       , sprite = 'entity/logistic-chest-requester'       },
  }



-- -------------------------------------------------------------------------- --
-- locallib                                                                   --
-- -------------------------------------------------------------------------- --
local Local = {}; const.Local = Local




-- "er:stockpile-passive-provider-7x19"
-- "er:stockpile-container-7x19"
function Local.encode_name(logistic_mode, width, height)
  height = height or width
  return (const.name.entity_prefix..'-%s-%sx%s'):format(logistic_mode, width, height)
  end

function Local.decode_name(name)
  -- assert(name:find(const.name.entity_prefix), 'Not a stockpile name.')
  local m, w, h
    = assert(name):match('^'..const.name.entity_prefix..'%-([%a%-]-)%-(%d+)x(%d+)$')
  return assert(m), assert(tonumber(w)), assert(tonumber(h))
  end

  
  do local function f(key) return function(logistic_mode)
    return assert(const.stockpile.logistic_modes[logistic_mode][key].logistic_mode)
    end end
Local.get_next_logistic_mode = f('next')
Local.get_prev_logistic_mode = f('prev')
  end
  
-- -------------------------------------------------------------------------- --
-- Const (dynamically derived)                                                --
-- -------------------------------------------------------------------------- --

const.name.entity_default = Local.encode_name(
  const.stockpile.default_logistic_mode,
  const.stockpile.default_width,
  const.stockpile.default_height
  )


-- This list is the *authorative* list of stockpile variations,
-- it is used for generating prototypes in data stages and 
-- as LuaGuiElement.anchor.names.
--
const.stockpile.names_array = (function(r)
  for logistic_mode in pairs(const.stockpile.logistic_modes) do
    for width = const.stockpile.min_width, const.stockpile.max_width do
      for height = const.stockpile.min_height, const.stockpile.max_height do
        -- containers can not be rotated.
        if (width == height)
        or (width * height <= const.stockpile.max_area)
        then
          local name = Local.encode_name(logistic_mode, width, height)
          table.insert(r, name)
          -- Test decoder.
          local m, w, h = Local.decode_name(name)
          if (m ~= logistic_mode) or (w ~= width) or (h ~= height) then
            print('Original    :', logistic_mode, width, height)
            print('De(Encode()):', m, w, h)
            error('Stockpile name de/encode loop is inconsistent. '
               .. 'See console for details.')
            end
          end
        end
      end
    end
 log('Will generate '.. #r ..' stockpile entity prototypes. This can take a while.') -- default: 2976
 return r end){}
  
do
  local _stockpile_names = (function(r,o)
    for _, name in ipairs(o) do r[name] = true end
    return r end)({}, const.stockpile.names_array)
function Local.is_stockpile_name(name)
  return (not not name) and (true == _stockpile_names[name]) end
  end
  
function Local.is_stockpile_spec(logistic_mode, width, height)
  return Local.is_stockpile_name(
    Local.encode_name(logistic_mode, width, height)
    )
  end
  
  
-- function Local.is_equal_stockpile(a, b)
--   local index = const.index
--   return (a and b)
--      and a[index.lu.width        ] == b[index.lu.width        ]
--      and a[index.lu.height       ] == b[index.lu.height       ]
--      and a[index.lu.logistic_mode] == b[index.lu.logistic_mode]
--   end
  
  
return const