﻿-- (c) eradicator a.k.a lossycrypt, 2017-2021, not seperately licensable
-- -------------------------------------------------------------------------- --

--[[ Notes:
  ]]

--[[ Annecdotes:
  ]]

--[[ Future:
  ]]
  
--[[ Todo:

  + Store Item content in entity tags when destroyed/mined.

  + Render Optimization:
    + Only update stockpiles that are visible to a player
      -> Instant/quick update selected / open
      -> Visibility is player.position
        -> zoom-to-world map mode is undetectable
          -> quick fix : selected.position
          -> proper fix: hijack map-mode button and use cursor-tracker
      
  + Number formatting
    + NoBreakSPace '\u{00A0}'
    + NARROW NO-BREAK SPACE '\u{202F}'
    + THIN SPACE '\u{2009}'
    + FIGURE SPACE '\u{2007}' ("preferred space for numbers")

      
  ]]
  
-- -------------------------------------------------------------------------- --
-- Built-In                                                                   --
-- -------------------------------------------------------------------------- --
local elroot = (pcall(require,'erlib/empty')) and '' or '__eradicators-library__/'
local say,warn,err,elreq,flag,ercfg=table.unpack(require(elroot..'erlib/shared'))

-- -------------------------------------------------------------------------- --
-- Eradicators Library                                                        --
-- (Factorio does not allow runtime require!)                                 --
-- -------------------------------------------------------------------------- --
local log         = elreq('erlib/lua/Log'          )().Logger  'stockpile-2'
-- local stop        = elreq('erlib/lua/Error'        )().Stopper 'stockpile-2'
-- local assertify   = elreq('erlib/lua/Error'        )().Asserter(stop)

-- local Verificate  = elreq('erlib/lua/Verificate'   )()
-- local verify      = Verificate.verify
-- local isType      = Verificate.isType

local Table       = elreq('erlib/lua/Table'        )()
-- local Array       = elreq('erlib/lua/Array'        )()
local Set         = elreq('erlib/lua/Set'          )()
-- local Filter      = elreq('erlib/lua/Filter'       )()
-- local Vector      = elreq('erlib/lua/Vector'       )()
local Math        = elreq('erlib/lua/Math'         )()
local Class       = elreq('erlib/lua/Class'        )()
-- local L           = elreq('erlib/lua/Lambda'       )()

-- local ntuples     = elreq('erlib/lua/Iter/ntuples' )()
-- local dpairs      = elreq('erlib/lua/Iter/dpairs'  )()
-- local sriapi      = elreq('erlib/lua/Iter/sriapi'  )()

-- local Setting     = elreq('erlib/factorio/Setting'   )()
local Gui         = elreq('erlib/factorio/Gui'       )()
local Player      = elreq('erlib/factorio/Player'    )()
local getp        = Player.get_event_player

-- -------------------------------------------------------------------------- --
-- Constants                                                                  --
-- -------------------------------------------------------------------------- --
local script = EventManager .get_managed_script    'stockpile-2'
local import = PluginManager.make_relative_require 'stockpile-2'
local const  = import '/const'
local Local  = const.Local
local index  = const.index

local StockpileGui   = import '/control/StockpileGui'
local Stockpile      = import '/control/Stockpile'
local InfinityTicker = import '/control/InfinityTicker'

-- -------------------------------------------------------------------------- --
-- Module                                                                     --
-- -------------------------------------------------------------------------- --

-- -------------------------------------------------------------------------- --
-- Local Library                                                              --
-- -------------------------------------------------------------------------- --

-- -------------------------------------------------------------------------- --
-- boolean

local function is_cursor_stockpile_item(cs)
  return not not (cs and cs.valid_for_read and cs.name == const.name.item)
  end
  
local function is_cursor_stockpile_blueprint(cs)
  return not not (cs and cs.valid_for_read and cs.name == const.name.blueprint)
  end

-- -------------------------------------------------------------------------- --
-- pdata
  
local function set_last_used_from_name(pdata, name)
  name = name or const.name.entity_default -- robust
  pdata.last_used = { Local.decode_name(name) }
  end

-- -------------------------------------------------------------------------- --
-- cursor_stack blueprint
 
local function update_cursor_blueprint(pdata, cs)
  assert(is_cursor_stockpile_blueprint(cs)) -- paranoia
  local m,w,h = table.unpack(pdata.last_used)
  --
  cs.set_blueprint_entities{{
    name          = Local.encode_name(m,w,h),
    tags          = {},
    position      = {0,0},
    entity_number = 1,
    }}
  --
  cs.allow_manual_label_change = false
  cs.label = ('%s %sx%s'):format(m,w,h)
  end

local function set_cursor_blueprint(pdata, cs)
  if pdata.p.clear_cursor() then
    cs.set_stack(const.name.blueprint)
    update_cursor_blueprint(pdata, cs)
    end
  end
  
-- -------------------------------------------------------------------------- --
-- Savedata                                                                   --
-- -------------------------------------------------------------------------- --
local Savedata
local SavedataDefaults = {players = {}, stockpiles = {}, ticked_arrays = {}}
PluginManager.manage_savedata  ('stockpile-2', function(_) Savedata = _ end, SavedataDefaults)
PluginManager.manage_garbage   ('stockpile-2')
PluginManager.classify_savedata('stockpile-2', {

  init_pdata = function(self, pindex)
    local pdata = Table.set(self.players, {assert(pindex)}, {
      p = game.players[pindex],
      pindex = pindex, -- not used?
      --
      })
    set_last_used_from_name(pdata, const.name.entity_default)
    return pdata end,

  get_pdata = function(self, e, pindex)
    local pdata = assert(self.players[pindex or e.player_index])
    return pdata, pdata.p end,

  sget_pdata = function(self, e, pindex)
    local pdata = self.players[pindex or e.player_index]
            or self:init_pdata(pindex or e.player_index)
    return pdata, pdata.p end,

  del_pdata = function(self, e, pindex)
    self.players[pindex or e.player_index] = nil
    end,
  
  })
  
-- -------------------------------------------------------------------------- --
-- Events                                                                     --
-- -------------------------------------------------------------------------- --
 

-- -------------------------------------------------------------------------- --
-- cursor_stack blueprint (size, logistic_mode)
  
local diffs = {
  -- both
  on_linked_input_cycle_blueprint_backwards     = {w = -1, h = -1},
  on_linked_input_cycle_blueprint_forwards      = {w =  1, h =  1},
  -- height only
  on_linked_input_smaller_terrain_building_area = {w =  0, h = -1},
  on_linked_input_larger_terrain_building_area  = {w =  0, h =  1},
  }
local function limit_width(width)
  return Math.limit_range(const.stockpile.min_width, width, const.stockpile.max_width)
  end
local function limit_height(height)
  return Math.limit_range(const.stockpile.min_height, height, const.stockpile.max_height)
  end
script.on_event(
  Table.keys(diffs)
  ,
  function(e)
    local pdata, p = Savedata:sget_pdata(e)
    local cs = p.cursor_stack
    if is_cursor_stockpile_blueprint(cs) then
      local diff = assert(diffs[e.input_name])
      local lu = pdata.last_used
      local w = limit_width(lu[index.lu.width ]  + diff.w)
      local h = limit_width(lu[index.lu.height]  + diff.h)
      if Local.is_stockpile_spec(lu[index.lu.logistic_mode], w, h) then
        lu[index.lu.width ], lu[index.lu.height] = w, h
        update_cursor_blueprint(pdata, cs)
        end
      end
    end)
    
script.on_event({
  'on_linked_input_rotate',
  'on_linked_input_reverse_rotate',
  },
  function(e)
    local pdata, p = Savedata:sget_pdata(e)
    --
    local cs = p.cursor_stack
    if is_cursor_stockpile_blueprint(cs) then
      local lu = pdata.last_used
      lu[index.lu.width ], lu[index.lu.height]   
      = lu[index.lu.height], lu[index.lu.width ] 
      update_cursor_blueprint(pdata, cs)
      return end
    --
    local selected = p.selected
    if selected and Local.is_stockpile_name(selected.name) then
      local pile = Stockpile(selected)
      local m = pile:decode_name()
      pile:set_logistic_mode(
        pdata
        ,
        (e.input_name == 'on_linked_input_rotate')
        and Local.get_next_logistic_mode(pile:decode_name())
         or Local.get_prev_logistic_mode(pile:decode_name())
        )
      return end
    end)
  
script.on_event(
  'on_linked_input_mine'
  ,
  function(e)
    local pdata, p = Savedata:sget_pdata(e)
    local cs = p.cursor_stack
    if is_cursor_stockpile_blueprint(cs) then
      local lu = pdata.last_used
      lu[index.lu.logistic_mode]
        = Local.get_next_logistic_mode(lu[index.lu.logistic_mode])
      update_cursor_blueprint(pdata, cs)
      end
    end)

  
-- -------------------------------------------------------------------------- --
-- item pickup
  
script.on_event(
  defines.events.on_player_pipette
  , 
  function(e)
    if e.item.name == const.name.item then -- was given correct item by pipette
      local pdata, p = Savedata:sget_pdata(e)
      local selected = p.selected
      local name = (selected.type == 'entity-ghost') and selected.ghost_name or selected.name
      if Local.is_stockpile_name(name) then
        set_last_used_from_name(pdata, name)
        local cs = p.cursor_stack
        set_cursor_blueprint(pdata, cs)
        end
      end
    end)

script.on_event(
  defines.events.on_player_cursor_stack_changed
  ,
  function(e)
    local pdata, p = Savedata:sget_pdata(e)
    local cs = p.cursor_stack
    if is_cursor_stockpile_item(cs) then
      set_cursor_blueprint(pdata, cs)
      end
    end)
    
    
  
  

-- -------------------------------------------------------------------------- --
-- gui  
  
script.on_event(
  defines.events.on_gui_opened
  ,
  function(e)
    if e.gui_type == defines.gui_type.entity then
      local pdata, p = Savedata:sget_pdata(e)
      local opened = p.opened
      if Local.is_stockpile_name(opened.name) then
        -- local logistic_mode, size = Local.decode_name(p.opened.name)
        StockpileGui(pdata):update_sidepanel(Stockpile(opened))
        end
      end
    end)
  
script.on_event({
  defines.events.on_gui_click,
  defines.events.on_gui_elem_changed,
  },
  function(e)
    local tags = e.element.tags
    if tags.module_name == 'er:stockpile-2' then
      local f = StockpileGui[tags.method_name]
      assert(f, 'Invalid method_name tag!')
      local pdata, p = Savedata:get_pdata(e)
      e.pile = Stockpile(p.opened)
      assert(e.pile, 'Gui click but no entity opened!')
      local gui = StockpileGui(pdata)
      assert(e.element.valid, 'Gui destroyed for no reason!')
      return f(gui, e)
      end
    end)
    
  
-- -------------------------------------------------------------------------- --
-- settings paste

script.on_event(
  defines.events.on_entity_settings_pasted
  ,
  function(e)
    local pile   = Stockpile(e.destination)
    local source = Stockpile(e.source)
    if pile and source then
      pile:paste_settings(source, (Savedata:get_pdata(e)))
      end
    end)

-- -------------------------------------------------------------------------- --
-- panic
  
script.on_event(script.events.on_user_panic, function(e)
  local pdata, p = Savedata:sget_pdata(e)
  if p.admin then
    Stockpile.full_savedata_repair()
    p.print{e.calming_words, 'Stockpiles'}
    end
  end)
