﻿-- (c) eradicator a.k.a lossycrypt, 2017-2021, not seperately licensable
-- -------------------------------------------------------------------------- --

--[[ Notes:
  ]]

--[[ Annecdotes:
  ]]

--[[ Future:

  + Performance optimization:
    Reuse as many tables as possible (walls, rows, tints) to reduce ram
    usage and copy time.
    => Pre-generate shifted+tinted max-width rows of walls, then
       select sub-segments for stockpiles without having to copy them again.
    => Reuse main entity table too (Table.scopy)
    => aggregate all prototypes into a single call to data:extend

  ]]
  
--[[ Todo:
  ]]
  
-- -------------------------------------------------------------------------- --
-- Built-In                                                                   --
-- -------------------------------------------------------------------------- --
local elroot = (pcall(require,'erlib/empty')) and '' or '__eradicators-library__/'
local say,warn,err,elreq,flag,ercfg=table.unpack(require(elroot..'erlib/shared'))

-- -------------------------------------------------------------------------- --
-- Eradicators Library                                                        --
-- (Factorio does not allow runtime require!)                                 --
-- -------------------------------------------------------------------------- --
local log         = elreq('erlib/lua/Log'          )().Logger  'stockpile-2'
-- local stop        = elreq('erlib/lua/Error'        )().Stopper 'stockpile-2'
-- local assertify   = elreq('erlib/lua/Error'        )().Asserter(stop)

-- local Verificate  = elreq('erlib/lua/Verificate'   )()
-- local verify      = Verificate.verify
-- local isType      = Verificate.isType

local Table       = elreq('erlib/lua/Table'        )()
local Array       = elreq('erlib/lua/Array'        )()
-- local Set         = elreq('erlib/lua/Set'          )()
-- local Filter      = elreq('erlib/lua/Filter'       )()
-- local Vector      = elreq('erlib/lua/Vector'       )()
local Color       = elreq('erlib/lua/Color'        )()

local ntuples     = elreq('erlib/lua/Iter/ntuples' )()
-- local dpairs      = elreq('erlib/lua/Iter/dpairs'  )()
-- local sriapi      = elreq('erlib/lua/Iter/sriapi'  )()

-- local Setting     = elreq('erlib/factorio/Setting'   )()
local Data        = elreq('erlib/factorio/Data/!init')()

local pairs
    = pairs

-- -------------------------------------------------------------------------- --
-- Constants                                                                  --
-- -------------------------------------------------------------------------- --
local asset  = PluginManager.make_asset_getter('stockpile-2', 'eradicators-stockpile')
local import = PluginManager.make_relative_require 'stockpile-2'
local const  = import '/const'
local Local  = const.Local

-- -------------------------------------------------------------------------- --
-- Module                                                                     --
-- -------------------------------------------------------------------------- --
local StoneWall = assert(data.raw.wall['stone-wall'], 'Another mod has deleted the "stone-wall" prototype.')

-- -------------------------------------------------------------------------- --
-- Local Library                                                              --
-- -------------------------------------------------------------------------- --
  
local function scopy(tbl) -- Skip LuaObject check from Table.scopy
  local r = {}
  for k,v in pairs(tbl) do r[k] = v end
  return r end
  
-- -------------------------------------------------------------------------- --
-- Data                                                                       --
-- -------------------------------------------------------------------------- --

Data.SimpleLinkedInput(
   'larger-terrain-building-area'
  ,'smaller-terrain-building-area'
  
  ,'cycle-blueprint-forwards'
  ,'cycle-blueprint-backwards'
  
  , 'mine'
  
  , 'rotate'
  , 'reverse-rotate'
  
  -- 'smart-pipette'
  )

Data.Inscribe {
  type = "item-with-tags",
  name = const.name.item,
  
  icons = {Data.Sprite.format_icon(asset('/item-icon-64².png'))},
  
  localised_description = {
    '',
    {'item-description.'..const.name.item},
    '\n',
    {'item-control-description.'    ..const.name.item}
    },

  subgroup = "storage",
  order = "a[items]-k[stockpile]-0",
  
  place_result = const.name.entity_default,
  stack_size = 50
  }
  
Data.Inscribe {
  type = "recipe",
  name = const.name.recipe,
  enabled = false,
  ingredients = {{"stone-wall", 100}},
  result = const.name.item,
  }
  
Data.Recipe.copy_unlock_condition('stone-wall', const.name.recipe)
  

local blueprint = Data.Inscribe {
  name = const.name.blueprint,
  magic = {
    -- must be "blueprint" for get_blueprint_entities()
    parent_type = 'blueprint',
    parent_name = 'blueprint',
    -- parent_type = 'copy-paste-tool',
    -- parent_name = 'copy-paste-tool',
    },
  icons = {Data.Sprite.format_icon(asset('/item-icon-64².png'))},
  
  
  }
table.insert(blueprint.flags, 'only-in-cursor')
table.insert(blueprint.flags, 'hidden')
Array.unsorted_remove_value(blueprint.flags, 'draw-logistic-overlay')

  
-- -------------------------------------------------------------------------- --

local walls_per_tile = 2 -- Does not work correctly for numbers other than 2.

local function scale_shift(shift, scale)
  -- print(scale)
  -- print(serpent.line(shift))
  -- print(serpent.line({shift[1] * scale, shift[2] * scale}))
  return {shift[1] * scale, shift[2] * scale} end

local function extract_variations(sprite)
  assert(sprite)
  local line_length = (sprite.line_length or 1)
  local r = {}
  for count = 0, (sprite.variation_count or 1) - 1 do
    r[#r+1] = {
      filename = sprite.filename,
      priority = sprite.priority,
      width    = sprite.width   ,
      height   = sprite.height  ,
      scale    = (sprite.scale or 1) / walls_per_tile,
      shift    = scale_shift(sprite.shift or {0,0}, 1 / walls_per_tile),
      -- shift    = sprite.shift,
      x        = (sprite.x or 0) + sprite.width  * (count % line_length)        ,
      y        = (sprite.y or 0) + sprite.height * math.floor(count/line_length),
      draw_as_shadow = sprite.draw_as_shadow,
      }
    end
  return r end

local wpictures = assert(StoneWall.pictures)
  
local pics = {
  y = {
    wall   = extract_variations(wpictures.straight_vertical.layers[1]),
    shadow = extract_variations(wpictures.straight_vertical.layers[2]),
    },
  x = {
    wall   = extract_variations(wpictures.straight_horizontal.layers[1]),
    shadow = extract_variations(wpictures.straight_horizontal.layers[2]),
    },
  lt = {
    wall   = extract_variations(wpictures.corner_right_down.layers[1]),
    shadow = extract_variations(wpictures.corner_right_down.layers[2]),
    },
  rt = {
    wall   = extract_variations(wpictures.corner_left_down.layers[1]),
    shadow = extract_variations(wpictures.corner_left_down.layers[2]),
    },
  lb = {
    wall   = extract_variations(wpictures.ending_right.layers[1]),
    shadow = extract_variations(wpictures.ending_right.layers[2]),
    },
  rb = {
    wall   = extract_variations(wpictures.ending_left.layers[1]),
    shadow = extract_variations(wpictures.ending_left.layers[2]),
    },
  }

local function get_random(variations)
  return table.deepcopy(variations[math.random(#variations)]) end
   
local function add_shift(a, b)
  return {a[1] + b[1], a[2] + b[2]}
  end

local ground = {
  -- Future: reducing width to 256 would increase texture reusage and reduce VRAM.
  filename = "__base__/graphics/terrain/nuclear-ground.png",
  width  = 1024,
  -- width  = 256,
  height = 256,
  scale  = 1,
  -- hr_version = {
    -- filename = "__base__/graphics/terrain/hr-nuclear-ground.png",
    -- },
  }
  
  
local function make_stockpile_picture(tint, width, height)
  local r = {layers = {}}
  local layers = r.layers
  
  -- Adds a single sprite
  local f0 = function(sprite)
    layers[#layers+1] = sprite
    end
  
  -- Adds a single wall+shadow.
  local f1 = function(group, shift)
    local wall   = get_random(group.wall)
    wall.shift   = add_shift(wall.shift  , shift)
    wall.tint    = tint
    f0(wall)
    --
    local shadow = get_random(group.shadow)
    shadow.shift = add_shift(shadow.shift, shift)
    f0(shadow)
    end
    
  -- Builds edges
  local f2 = function(group, ix, iy)
    for x = ix[1], ix[2] or ix[1], ix[3] or 1 do
      for y = iy[1], iy[2] or iy[1], iy[3] or 1 do
        f1(group, {x, y})
        end
      end
    end
  
  -- These numbers are hand-crafted for 2-walls-per-tile.
  local a = 1 / walls_per_tile
  local b = a * 3/2
  local c = a * 1/2
  

  -- filling (nuclear ground has nice contrast)
  -- (Disabled: too slow)
  if false then
    local px_w = (width *32) - (c*32)
    local px_h = (height*32) - (c*32)
    
    local master_shift = {(px_w/2)/32, (px_h/2)/32}
    
    local row = {}
    repeat
      local this = {
        filename = ground.filename,
        width  = ground.width,
        height = ground.height,
        x = 0, y = 0,
        tint = Color.premultiply_alpha(Color('White'), 0.5),
        shift = {0,0}, -- adjusted later
        }
      row[#row+1] = this
      --
      px_w = px_w - ground.width
      if px_w < 0 then this.width = this.width + px_w end
      --
      local total_width = 0
      for i = 1, #row do
        local this = row[i]
        this.shift[1] = ((total_width + this.width/2) / 32) - master_shift[1]
        total_width = total_width + this.width
        end
      until px_w <= 0
      
    local rows = {row}
    
    repeat
      px_h = px_h - ground.height
      if px_h < 0 then
        local row = rows[#rows]
        for i, this in ipairs(row) do
          this.height = this.height + px_h
          end
        end
      if px_h > 0 then rows[#rows+1] = Table.dcopy(row) end
      until px_h <= 0
      
    local total_height = 0
    for _, row in ipairs(rows) do
      local dy = ((total_height + row[1].height/2) / 32) - master_shift[2]
      for _, this in ipairs(row) do
        this.shift[2] = dy
        f0(this)
        end
      total_height = total_height + row[1].height
      end
    end

      
  -- top corners
  f1(pics.lt, {-width/2 + a/2, -height/2 + a/2})
  f1(pics.rt, { width/2 - a/2, -height/2 + a/2})
  
  -- horizontal edges
  -- f2(pics.x, {-width/2 +b,  width/2 -b, a}, {-height/2 +(a*13/16)})
  f2(pics.x, {-width/2 +b,  width/2 -b, a}, {-height/2 +c})
  f2(pics.x, {-width/2 +b,  width/2 -b, a}, { height/2 -c})
  
  -- vertical edges
  -- f2(pics.y, {-width/2 +c}, {-height/2 +(a*5/4),  height/2 -b, a})
  -- f2(pics.y, { width/2 -c}, {-height/2 +(a*5/4),  height/2 -b, a})
  f2(pics.y, {-width/2 +c}, {-height/2 +b,  height/2 -b, a})
  f2(pics.y, { width/2 -c}, {-height/2 +b,  height/2 -b, a})

  -- bottom corners
  f1(pics.lb, {-width/2 + a/2,  height/2 - a/2})
  f1(pics.rb, { width/2 - a/2,  height/2 - a/2})
  

  
  return r end
  

-- -------------------------------------------------------------------------- --
local function make_stockpile_entity(name, logistic_mode, width, height, slot_count)
  
  local is_logistic = not (logistic_mode == 'container')
  
  local scale = 1 /( 640 / (width*32)) --unscaled picture width is 20.0
 
  local connector = circuit_connector_definitions.create(universal_connector_template,{{
    variation = 0,
    main_offset   = { 1*(width/2 - 0.1) - scale*0.5, 1*(height/2 - 0.1) - scale*1.5},
    shadow_offset = { 1*(width/2 - 0.1) - scale*0.5, 1*(height/2 - 0.1) - scale*1.5},
    show_shadow = true
    }})

  local wall_count = (width + height) * 2 -- number of circumfencing wall segments
    
  local entity = {
    name = name,
    type = is_logistic and 'logistic-container' or 'container',
    
    icons = {Data.Sprite.format_icon(asset('/item-icon-64².png'))},

    flags = {"placeable-neutral", "player-creation", "hide-alt-info"},
    minable = {mining_time = 0.75, result = const.name.item},
    max_health = wall_count * StoneWall.max_health,
    order = "a[items]-k[stockpile]",
    placeable_by = {item = const.name.item, count = 1},
    
    corpse = "small-remnants",
    open_sound  = { filename = "__base__/sound/metallic-chest-open.ogg" , volume = 0.65 },
    close_sound = { filename = "__base__/sound/metallic-chest-close.ogg", volume = 0.7  },
    
    resistances = table.deepcopy(StoneWall.resistances),
    
    fast_replaceable_group = "container",
    
    -- collision_mask = {"item-layer", "object-layer", "water-tile"}, -- not "player-layer"
    
    collision_box = {
      {-1*(width/2 - 0.24), -1*(height/2 - 0.24)},
      { 1*(width/2 - 0.24),  1*(height/2 - 0.24)}
      },
    selection_box = {
      -- {-1*(width/2 + 0.24), -1*(height/2 + 0.24)},
      -- { 1*(width/2 + 0.24),  1*(height/2 + 0.24)},
      {-1*(width/2 - 4/32), -1*(height/2 - 4/32)},
      { 1*(width/2 - 4/32),  1*(height/2 - 4/32)},
      },
      
    inventory_size = slot_count,
    
    logistic_mode = is_logistic and logistic_mode or nil,
    max_logistic_slots = (logistic_mode == 'storage') and 1 or nil,
    
    -- This causes HUGE ram usage and extremely slow startup.
    additional_pastable_entities = const.stockpile.names_array,
    
    circuit_wire_connection_point = connector.points,
    circuit_connector_sprites = connector.sprites,
    circuit_wire_max_distance = default_circuit_wire_max_distance,
    
    -- picture = { layers = {
    --   { priority = "extra-high",
    --     filename = asset '/640_wall.png', width = 640, height = 652,
    --     -- shift = util.by_pixel(32*0.5*width, 0),
    --     shift = {0.3*scale, 0.1*scale},
    --     scale = scale,
    --     },
    --   { priority = "extra-high",
    --     filename = asset '/640_shadow.png', width = 640, height = 652,
    --     -- shift = util.by_pixel(32*0.5*width, 0),
    --     shift = {0.3*scale, 0.1*scale},
    --     scale = scale,
    --     draw_as_shadow = true,
    --     }
    --   } },
    }
    
  entity.localised_name        = {'item-name.'..const.name.item}
  entity.localised_description = {'item-description.'..const.name.item}
       
       
  local tint = is_logistic and Color(logistic_mode) or nil
  if tint then tint.a = 1 end
  entity.picture = make_stockpile_picture(tint, width, height)
       
  -- if is_logistic then
    -- print(serpent.block(entity.picture))
    -- print(serpent.line(tint))
    -- error()
    -- end
       
  data:extend{entity}
  -- log:debug('Created Stockpile: ', entity.name)
  end

-- -------------------------------------------------------------------------- --

for _, name in ipairs(const.stockpile.names_array) do
  local logistic_mode, width, height = Local.decode_name(name)
  local slot_count = math.min(
    (width-1) * (height-1) * const.stockpile.slots_per_tile_intersection, 
    const.stockpile.maximum_slot_count
    )
  make_stockpile_entity(name, logistic_mode, width, height, slot_count)
  end
