﻿
-- local stack_sizes = Memoize(function(item_name)
  -- return game.item_prototypes[item_name].stack_size
  -- end)
  
local stack_sizes = Prototype.data_cache('item', 'stack_size')


local stack_sizes = Prototype.data_cache('item',
  function(x) if x.type == 'item' then return x.stack_size end end)

local function new_pile(entity, item_name)
  
  -- Do i care about the number? Probably not.
  local registration_number = script.register_on_entity_destroyed(entity)

  local inventory = entity.get_inventory(defines.inventory.chest)
  
  
  
  local pile = {
    container = entity,
    unit_number = entity.unit_number,
    
    -- inventory = inventory,
    
    -- can function references be stored in Savedata?
    --> doesn't matter! Just get the in on_load if not.
    -- get_item_count = inventory.get_item_count,
    --> managed to crash by saving after storing in global :p
    
    -- SimpleItemStack used as ItemStackIdentification
    -- Stored to prevent creating garbage tables. (GC optimization)
    simple_stack = {name = item_name, count = 0},
    
    -- Amount of items that would fill half of the inventory.
    half_count = (#inventory/2) * assert(stack_sizes[item_name], 'Not a normal item.'),
    
    -- Count of items virtualized.
    virtual_count = 0,
    
    }
    
  table.insert(Savedata.infinipiles, pile)
  Savedata.infinipile_lookup[pile.unit_number] = pile
  end


-- Cache inventory access functions on the lua side (performance optimization).  
local InventoryCache = Memoize(function(unit_number)
  local pile
  for _, this in ipairs(Savedata.infinipiles) do
    if this.unit_number == unit_number then pile = this break end
    end
  assert(pile)
  return {
    get_item_count = inventory.get_item_count,
    insert         = inventory.insert,
    remove         = inventory.remove,
    }
  end)


local function add_method_cache(infinite_inventory)
  -- Lua object methods can not be safely stored in global, 
  -- and LuaInventory does not have a unique index to use for caching.
  -- Meta-lookup should stil be faster than c-lookup.
  setmetatable(infinite_inventory, {
    __index = {
      get_item_count = infinite_inventory.inventory.get_item_count,
      insert         = infinite_inventory.inventory.insert        ,
      remove         = infinite_inventory.inventory.remove        ,
      }
    })
  end
  
scrip.on_load(function()
  for _, infinite_inventory in ipairs(Savedata.inventories) do
    add_method_cache(infinite_inventory)
    end
    
  -- ForEach(Savedata.inventories, add_method_cache)
    
  end)
  
  
-- TODO: Make this work on *any* given inventory and stay agnostic
-- to what the inventory is attached to.

-- Feature creep:
-- How much slower/complicated is it with N item types instead of 1?


-- The fastest possible update of one infinity stack:
-- At most 2(!) C-function calls (+ looking up the function).
-- (count & insert) or (count & remove)

scipt.on_event(defines.events.on_tick, function()
  local infinipiles = Savedata.infinipiles
  --
  for i=1, #infinipiles do
    local pile = infinipiles[i]
    local inventory = InventoryCache[pile.unit_number]
    local count = pile.get_item_count(pile.item_name)
    pile.total_count = pile.virtual_count + count
    --
    if count > pile.half_count then
      local simple_stack = pile.ItemStackIdentification
      pile.simple_stack.count = count - pile.half_count
      pile.virtual_count = pile.virtual_count + inventory.remove(pile.simple_stack)
      
    elseif count < pile.half_count then
      local simple_stack = pile.ItemStackIdentification
      pile.simple_stack.count = pile.half_count - count
      pile.virtual_count = pile.virtual_count - inventory.insert(pile.simple_stack)
      end
    --
    if pile.empty_alert and count == 0 then
      spawn_alert(pile.container.position, pile.item_name)
      end
    --
    if pile.show_number or true then
      update_number(pile)
      end 
    end
  end)
  
-- Api Doc:
-- "Note: Depending on when a given entity is destroyed this will
--  be fired at the end of the current tick or end of the next tick."
script.on_event(defines.events.on_entity_destroyed, function(e)
  
  local pile = Savedata.infinipile_lookup[e.unit_number]
  if pile then
    Savedata.infinipile_lookup[pile.unit_number] = nil
    for i, this in ipairs(Savedata.infinipiles) do
      
      end
    end
  end)
  
  
script.on_event(EventManager.events.on_entity_created, function(e)
  if e.created_entity.name == 'infinipile' then
    new_pile(e.created_entity)
    
    end
  end)